-- VehicleSort.lua for FS19
-- Author: sperrgebiet
-- Please see https://github.com/sperrgebiet/FS19_VehicleExplorer for additional information, credits, issues and everything else

VehicleSort = {};
VehicleSort.eventName = {};

VehicleSort.ModName = g_currentModName;
VehicleSort.ModDirectory = g_currentModDirectory;
VehicleSort.Version = "0.9.1.0";


VehicleSort.debug = fileExists(VehicleSort.ModDirectory ..'debug');

print(string.format('VehicleSort v%s - DebugMode %s)', VehicleSort.Version, tostring(VehicleSort.debug)));

VehicleSort.bgTransDef = 0.8;
VehicleSort.txtSizeDef = 2;
VehicleSort.infoYStart = 0.8;
VehicleSort.listAlignment = 2;						-- 1 = Left, 2 = Center, 3 = Right)
VehicleSort.showImgMaxImp = 9;
VehicleSort.showInfoMaxImpl = 9;
VehicleSort.showImplementsMax = 3;

-- Integration environment for Tardis
envTardis = nil;

VehicleSort.config = {											--Id		-Order in configMenu
  {'showTrain', true, 1},										-- 1		1
  {'showCrane', false, 2},                             			-- 2		2
  {'showBrand', false, 3},                             			-- 3		3
  {'showHorsepower', true, 4},                         			-- 4		4
  {'showNames', true, 8},                              			-- 5		8
  {'showFillLevels', true, 5},                         			-- 6		5
  {'showPercentages', true, 6},                        			-- 7		6
  {'showEmpty', false, 7},                             			-- 8		7
  {'txtSize', VehicleSort.txtSizeDef, 16},             			-- 9		16
  {'bgTrans', VehicleSort.bgTransDef, 17},              		-- 10		17
  {'showSteerableImplements', true, 11},                		-- 11		11
  {'showImplements', true, 9},	                         		-- 12		9
  {'showHelp', true, 27},                               		-- 13		27
  {'saveStatus', true, 24},                             		-- 14		24
  {'showImg', true, 12},                                		-- 15		12
  {'showInfo', true, 14},                               		-- 16		14
  {'infoStart', VehicleSort.infoYStart, 21},            		-- 17		21
  {'infoBg', true, 19},                                 		-- 18		19
  {'imageBg', true, 20},                                		-- 19		20
  {'listAlignment', VehicleSort.listAlignment, 22},     		-- 20		22
  {'cleanOnRepair', true, 23},                          		-- 21		23
  {'integrateTardis', true, 25},                        		-- 22		25
  {'enterVehonTeleport', true, 26},                     		-- 23		26
  {'showImgMaxImp', VehicleSort.showImgMaxImp, 13},     		-- 24		13
  {'showInfoMaxImpl', VehicleSort.showInfoMaxImpl, 15},			-- 25		15
  {'showImplementsMax', VehicleSort.showImplementsMax, 10},		-- 26		10
  {'useTwoColoredList', true, 18},								-- 27		18
};

VehicleSort.tColor = {}; -- text colours
VehicleSort.tColor.isParked 	= {0.5, 0.5, 0.5, 0.7};   -- grey
VehicleSort.tColor.locked 		= {1.0, 0.0, 0.0, 1.0};   -- red
VehicleSort.tColor.selected 	= {0.8879, 0.1878, 0.0037, 1.0}; -- orange
VehicleSort.tColor.standard 	= {1.0, 1.0, 1.0, 1.0}; -- white
VehicleSort.tColor.standard2 	= {0.8228, 0.8388, 0.7304, 1.0}; -- eggcolor
VehicleSort.tColor.hired 		= {0.0, 0.5, 1.0, 1.0}; 	-- blue
VehicleSort.tColor.followme 	= {0.92, 0.31, 0.69, 1.0}; 	-- light pink
VehicleSort.tColor.self  		= {0.0, 1.0, 0.0, 1.0}; -- green
VehicleSort.tColor.motorOn		= {0.9301, 0.7605, 0.0232, 1.0}; -- yellow

VehicleSort.keyCon = 'VeExConfig';
VehicleSort.selectedConfigIndex = 1;
VehicleSort.selectedRealConfigIndex = 1;
VehicleSort.selectedIndex = 1;
VehicleSort.selectedLock = false;
VehicleSort.showConfig = false;
VehicleSort.showSteerables = false;
VehicleSort.xmlAttrId = '#vsid';
VehicleSort.xmlAttrOrder = '#vsorder';
VehicleSort.xmlAttrParked = '#vsparked';
VehicleSort.Sorted = {};
VehicleSort.HiddenCount = 0;
--VehicleSort.dirtyState = false;						-- Used to check if we have to sync the order with g_currentMission.vehicles
VehicleSort.orderedConfig = {};							-- It's just nicer to have the config list ordered

addModEventListener(VehicleSort);

function VehicleSort:dp(val, fun, msg) -- debug mode, write to log
  if not VehicleSort.debug then
    return;
  end
  if msg == nil then
    msg = ' ';
  else
    msg = string.format(' msg = [%s] ', tostring(msg));
  end
  local pre = 'VehicleSort DEBUG:';
  if type(val) == 'table' then
    if #val > 0 then
      print(string.format('%s BEGIN Printing table data: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
      DebugUtil.printTableRecursively(val, '.', 0, 3);
      print(string.format('%s END Printing table data: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
    else
      print(string.format('%s Table is empty: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
    end
  else
    print(string.format('%s [%s]%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
  end
end


function VehicleSort:prerequisitesPresent(specializations)
	return true;
end

function VehicleSort:loadMap(name)
	print("--- loading VehicleSort V".. VehicleSort.Version .. " | ModName " .. VehicleSort.ModName .. " ---");
	FSBaseMission.registerActionEvents = Utils.appendedFunction(FSBaseMission.registerActionEvents, VehicleSort.RegisterActionEvents);
	
	VehicleSort:initVS();
	VehicleSort:loadConfig();
	
	-- We'd get an error when we want to access the config menu and the data is not populated yet. Same as there is some functionality in corner cases where we
	-- already need a populated Sorted list, hence we're going to pouplate it once here
	for i = 1, #VehicleSort.config do
		local val = {i, VehicleSort.config[i]};
		table.insert(VehicleSort.orderedConfig, val);
	end
	table.sort(VehicleSort.orderedConfig, function(a, b) return a[2][3] < b[2][3] end)	
	
	VehicleSort.Sorted = VehicleSort:getOrderedVehicles();
end

function VehicleSort:onLoad(savegame)
end

function VehicleSort:onPostLoad(savegame)

	if savegame ~= nil then
		local xmlFile = savegame.xmlFile;
		local key = savegame.key..".vehicleSort";
		
		local orderId = getXMLInt(xmlFile, key.."#UserOrder");
		if orderId ~= nil then
			VehicleSort:dp(string.format('Loaded orderId {%d} for vehicleId {%d}', orderId, self.id), 'onPostLoad');
		end
		
		if self.spec_vehicleSort ~= nil then
			local specVS = self.spec_vehicleSort;
			specVS.id = self.id;
			if orderId ~= nil then
				specVS.orderId = orderId;
			end
		end
		
		local isParked = Utils.getNoNil(getXMLBool(xmlFile, key.."#isParked"), false);
		if isParked then
			VehicleSort:dp(string.format('Set isParked {%s} for orderId {%d} / vehicleId {%d}', tostring(isParked), orderId, self.id), 'onPostLoad');
			self:setIsTabbable(false);
		end
		
		--Check to avoid issues after a vehicle reset
		if VehicleSort.Sorted[orderId] ~= nil then
			if g_currentMission.vehicles[VehicleSort.Sorted[orderId]]['id'] ~= self.id then
				VehicleSort:dp('Sync required after vehicle reset');
				VehicleSort:SyncSorted();
			end
		end
	end
end

function VehicleSort:RegisterActionEvents(isSelected, isOnActiveVehicle)

	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsToggleList',self, VehicleSort.action_vsToggleList ,false ,true ,false ,true)
	if result then
		table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end

	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsLockListItem',self, VehicleSort.action_vsLockListItem ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsMoveCursorUp',self, VehicleSort.action_vsMoveCursorUp ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsMoveCursorDown',self, VehicleSort.action_vsMoveCursorDown ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsMoveCursorUpFast',self, VehicleSort.action_vsMoveCursorUpFast ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsMoveCursorDownFast',self, VehicleSort.action_vsMoveCursorDownFast ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsChangeVehicle',self, VehicleSort.action_vsChangeVehicle ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsShowConfig',self, VehicleSort.action_vsShowConfig ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end	
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsTogglePark',self, VehicleSort.action_vsTogglePark ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end	

	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsRepair',self, VehicleSort.action_vsRepair ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsTab',self, VehicleSort.action_vsTab ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
	local result, eventName = InputBinding.registerActionEvent(g_inputBinding, 'vsTabBack',self, VehicleSort.action_vsTabBack ,false ,true ,false ,true)
	if result then
        table.insert(VehicleSort.eventName, eventName);
		g_inputBinding.events[eventName].displayIsVisible = VehicleSort.config[13][2];
    end
	
end

function VehicleSort.registerEventListeners(vehicleType)
	local functionNames = {	"onLoad", "onPostLoad", "saveToXMLFile" };
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, VehicleSort);
	end
end

function VehicleSort:keyEvent(unicode, sym, modifier, isDown)
end

function VehicleSort:mouseEvent(posX, posY, isDown, isUp, button)
	if VehicleSort:isActionAllowed() and ( isDown and button == Input.MOUSE_BUTTON_LEFT) then
		VehicleSort.action_vsChangeVehicle();
	end

	if VehicleSort:isActionAllowed() and ( isDown and button == Input.MOUSE_BUTTON_RIGHT) then
		VehicleSort.action_vsLockListItem();
	end

	if VehicleSort:isActionAllowed() and ( isDown and Input.isMouseButtonPressed(Input.MOUSE_BUTTON_WHEEL_UP)) then
		VehicleSort.action_vsMoveCursorUp();
	end
	
	if VehicleSort:isActionAllowed() and ( isDown and Input.isMouseButtonPressed(Input.MOUSE_BUTTON_WHEEL_DOWN)) then
		VehicleSort.action_vsMoveCursorDown();
	end	
end

function VehicleSort:saveToXMLFile(xmlFile, key)
	VehicleSort:dp(string.format('key {%s}', key), 'saveToXMLFile');
	if self.spec_vehicleSort ~= nil then
		if self.spec_vehicleSort.orderId ~= nil then
			setXMLInt(xmlFile, key.."#UserOrder", self.spec_vehicleSort.orderId);
		end
		
		if VehicleSort:isParked(self.spec_vehicleSort.realId) then
			setXMLBool(xmlFile, key.."#isParked", true);
		end
	end
end

function VehicleSort:draw()
	
	--VehicleSort:dp(string.format('showConfig [%s] & showSteerables [%s]', tostring(VehicleSort.showConfig), tostring(VehicleSort.showSteerables)));
  
	if VehicleSort.showConfig or VehicleSort.showSteerables then
		local dbgY = VehicleSort.dbgY;
		VehicleSort.bgY = nil;
		VehicleSort.bgW = nil;
		VehicleSort.bgH = nil;
		if VehicleSort.showConfig then
		  VehicleSort:drawConfig();
		else
		  VehicleSort:drawList();
		end
	end
	
end

function VehicleSort:delete()
end

function VehicleSort:deleteMap()
end

-- Functions for actionEvents/inputBindings

function VehicleSort:action_vsToggleList(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("vsToggleList fires", "vsToggleList");

	if envTardis == nil and VehicleSort.config[22][2] then
		-- Integration with Tardis
		local TardisName = "FS19_Tardis";

		if g_modIsLoaded[TardisName] then
			envTardis = getfenv(0)[TardisName];
			print("VehicleExplorer: Tardis integration available");
		end
	end
		
	if VehicleSort.showSteerables and not VehicleSort.showConfig then
		VehicleSort.showSteerables = false;
		VehicleSort.selectedLock = false;
		
		-- When dirtyState is true it means there was a resort going on, hence we save the ordered list back to g_currentMission.vehicles to have a proper 'tab'-order
		if VehicleSort.dirtyState then
			-- I don't like to alter g_currentMission.vehicles, as I have no idea what other mods do with it, and actually it doesn't help with the tab order anyways
			-- VehicleSort:SyncSortedWithGame();
		end
	else
		VehicleSort.showSteerables = true;
		if VehicleSort.showConfig then
			VehicleSort.saveConfig();
		end
		VehicleSort.showConfig = false;
    end
end

function VehicleSort:action_vsLockListItem(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("vsLockListItem fires", "vsLockListItem");
	if VehicleSort.showSteerables then
		if not VehicleSort.selectedLock and VehicleSort.selectedIndex > 0 then
			VehicleSort.selectedLock = true;
		elseif VehicleSort.selectedLock then
			VehicleSort.selectedLock = false;
		end
	elseif VehicleSort.showConfig then
		if VehicleSort:contains({9, 20}, VehicleSort.selectedRealConfigIndex) then
			VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] + 1;
			if VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] > 3 then
				VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = 1;
			end
		elseif VehicleSort:contains({24, 25, 26}, VehicleSort.selectedRealConfigIndex) then
			VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] + 1;
			if VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] > 9 then
				VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = 1;
			end			
		elseif VehicleSort:contains({10, 17}, VehicleSort.selectedRealConfigIndex) then
			VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] + 0.1;
			if VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] > 1 then
				VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = 0.0;
			end
		else
			VehicleSort.config[VehicleSort.selectedRealConfigIndex][2] = not VehicleSort.config[VehicleSort.selectedRealConfigIndex][2];
		end
	end
end

function VehicleSort:action_vsMoveCursorUp(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsMoveCursorUp fires", "action_vsMoveCursorUp");
	if VehicleSort.showSteerables then
		if Input.isKeyPressed(KEY_lalt) then
			VehicleSort:moveUp(3);
		else
			VehicleSort:moveUp(1);
		end
	elseif VehicleSort.showConfig then
		VehicleSort:moveConfigUp();
	end
end

function VehicleSort:action_vsMoveCursorDown(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsMoveCursorDown fires", "action_vsMoveCursorDown");
	if VehicleSort.showSteerables then
		VehicleSort:moveDown(1);
	elseif VehicleSort.showConfig then
		VehicleSort:moveConfigDown();
	end	
end

function VehicleSort:action_vsMoveCursorUpFast(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsMoveCursorUpFast fires", "action_vsMoveCursorUpFast");
	if VehicleSort.showSteerables then 
		VehicleSort:moveUp(3);
	end	
end

function VehicleSort:action_vsMoveCursorDownFast(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsMoveCursorDownFast fires", "action_vsMoveCursorDownFast");
	if VehicleSort.showSteerables then 
		VehicleSort:moveDown(3);
	end	
end

function VehicleSort:action_vsChangeVehicle(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsChangeVehicle fires", "action_vsChangeVehicle");
	if VehicleSort.showSteerables then
		local realVeh = g_currentMission.vehicles[VehicleSort.Sorted[VehicleSort.selectedIndex]];
		if not realVeh:getIsControlled() then

			VehicleSort:dp(string.format('VehicleSort.wasTeleportAction {%s}', tostring(VehicleSort.wasTeleportAction)));
			
			if envTardis == nil or 
						(envTardis ~= nil and not VehicleSort.wasTeleportAction) or 
						(envTardis ~= nil and VehicleSort.wasTeleportAction and VehicleSort.config[23][2]) then
				g_currentMission:requestToEnterVehicle(realVeh);
				VehicleSort.wasTeleportAction = false;
			elseif envTardis ~= nil then
				VehicleSort.wasTeleportAction = false;
			end
		end
	end
end

function VehicleSort:action_vsShowConfig(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsShowConfig fires", "action_vsShowConfig");
	if VehicleSort.showSteerables and not VehicleSort.showConfig then
      VehicleSort.showSteerables = false;
	end
    
	VehicleSort.showConfig = not VehicleSort.showConfig;
	VehicleSort:saveConfig();
end

function VehicleSort:action_vsTogglePark(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp("action_vsTogglePark fires", "action_vsTogglePark");
	if VehicleSort.showSteerables then
		VehicleSort:toggleParkState(VehicleSort.selectedIndex);
	end
end

function VehicleSort:action_vsRepair(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp(string.format('action_vsRepair fires - VehicleSort.showSteerables {%s}', tostring(VehicleSort.showSteerables)), "action_vsRepair");
	if VehicleSort.showSteerables then
	
		local infoText = "";
		VehicleStatus:RepairVehicleWithImplements(VehicleSort.Sorted[VehicleSort.selectedIndex]);
		infoText = g_i18n.modEnvironments[VehicleSort.ModName].texts.RepairDone;
		
		if VehicleSort.config[21][2] then
			VehicleStatus:CleanVehicleWithImplements(VehicleSort.Sorted[VehicleSort.selectedIndex]);
			infoText = g_i18n.modEnvironments[VehicleSort.ModName].texts.RepairCleaningDone;
		end
		
		g_currentMission:showBlinkingWarning(infoText, 2000);
		
	end
end

function VehicleSort:action_vsTab(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp(string.format('action_vsTab fires - VehicleSort.showSteerables {%s}', tostring(VehicleSort.showSteerables)), "action_vsTab");
	VehicleSort:tabVehicle();
end

function VehicleSort:action_vsTabBack(actionName, keyStatus, arg3, arg4, arg5)
	VehicleSort:dp(string.format('action_vsTabBack fires - VehicleSort.showSteerables {%s}', tostring(VehicleSort.showSteerables)), "action_vsTabBack");
	VehicleSort:tabVehicle(true);
end

--
-- VehicleSort specific functions
--
function VehicleSort:calcPercentage(curVal, maxVal)
	local per = curVal / maxVal * 100;
	return (math.floor(per * 10)/10);
end

function VehicleSort:drawConfig()
	local xPos = VehicleSort.tPos.x;
	local yPos = VehicleSort.tPos.y;
	local size = VehicleSort:getTextSize();
	local y = VehicleSort.tPos.y;
	local txtOn = g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_on;
	local txtOff = g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_off;
	local texts = {};
	VehicleSort.bgW = (VehicleSort.tPos.columnWidth * 2 )+ VehicleSort.tPos.padSides + getTextWidth(size, txtOff);		--For config we can use wider columns


		-- We render the heading seperately to have it centered, despite the user config for the list
	local headingY = VehicleSort.tPos.y + size + (VehicleSort.tPos.padHeight * 6);
	local txt = g_i18n.modEnvironments[VehicleSort.ModName].texts.configHeadline;
	setTextAlignment(VehicleSort.tPos.alignmentC);
	setTextColor(unpack(VehicleSort.tColor.standard));
	renderText(VehicleSort.tPos.x, headingY, size + VehicleSort.tPos.sizeIncr, tostring(txt)); -- x, y, size, txt

	setTextAlignment(VehicleSort.tPos.alignmentL);
	
	--VehicleSort:dp(orderedConfig, 'drawConfig');
	-- And now the rest of our config
	for k, v in ipairs(VehicleSort.orderedConfig) do
		local clr = VehicleSort.tColor.standard;
		if k == VehicleSort.selectedConfigIndex then
		  clr = VehicleSort.tColor.selected;
		end
		local rText = g_i18n.modEnvironments[VehicleSort.ModName].texts[VehicleSort.config[v[1]][1]];
		local state = VehicleSort.config[v[1]][2];
		if VehicleSort:contains({9, 24, 25, 26}, v[1]) then							--txtSize, showImgMaxImp, showInfoMaxImpl, showImplementsMax as int
			state = string.format('%d', state);
		elseif VehicleSort:contains({10, 17}, v[1]) then						--bgTransparency, VehicleSort.infoYStart as float
			state = string.format('%.1f', state);
		elseif v[1] == 20 then 			-- List text alignment
			if state == 1 then
				state = g_i18n.modEnvironments[VehicleSort.ModName].texts.left;
			elseif state == 2 then
				state = g_i18n.modEnvironments[VehicleSort.ModName].texts.center;
			elseif state == 3 then
				state = g_i18n.modEnvironments[VehicleSort.ModName].texts.right;
			end
		elseif state then
		  state = txtOn;
		else
		  state = txtOff;
		end
		table.insert(texts, {xPos - (VehicleSort.bgW / 2) + VehicleSort.tPos.padSides, yPos, size, clr, rText}); --config definition line
		table.insert(texts, {xPos - (VehicleSort.bgW / 2) + (VehicleSort.tPos.columnWidth * 2), yPos, size, clr, state}); --config value
		yPos = yPos - size - VehicleSort.tPos.spacing;
	end
  
	VehicleSort.bgY = yPos;
	VehicleSort.bgH = (y - yPos) + size + VehicleSort.tPos.yOffset + VehicleSort.tPos.padHeight;
	if VehicleSort.bgY ~= nil and VehicleSort.bgW ~=nil and VehicleSort.bgH ~= nil then
		VehicleSort:renderBg(VehicleSort.bgX, VehicleSort.bgY, VehicleSort.bgW, VehicleSort.bgH);
	end;  

	setTextBold(false);
	for k, v in ipairs(texts) do
		setTextColor(unpack(v[4]))
		renderText(v[1], v[2], v[3], tostring(v[5]));
	end
	setTextColor(unpack(VehicleSort.tColor.standard));
  
	--Show the last selected vehicle for info/image position & BG option
	if VehicleSort:contains({17, 18, 19, 24, 25}, VehicleSort.selectedRealConfigIndex) then
		if g_currentMission.vehicles[VehicleSort.Sorted[VehicleSort.selectedIndex]] ~= nil then
			VehicleSort:drawInfobox(VehicleSort.Sorted[VehicleSort.selectedIndex]);
			VehicleSort:drawStoreImage(VehicleSort.Sorted[VehicleSort.selectedIndex]);
		end
	end
  
end

function VehicleSort:drawList()
  VehicleSort.Sorted = VehicleSort:getOrderedVehicles();
   
   if VehicleSort.HiddenCount == #VehicleSort.Sorted then
		VehicleSort:showNoVehicles();
		VehicleSort.showSteerables = false;
		return false;
   end
   
  --VehicleSort:dp(vehList, 'drawList', 'vehList');
  
	local cnt = #VehicleSort.Sorted;
	if cnt == 0 then
		return;
	end
	setTextBold(true); -- for width checks, to compensate for increased width when the line is bold
 
	local yPos = VehicleSort.tPos.y;
	local bgPosY = yPos; 
	local size = VehicleSort.getTextSize();
	local y = VehicleSort.tPos.y;
	local txt = g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_title;
	local texts = {};
	local bold = false;
	local minBgW = 0;
	VehicleSort.bgY = y - VehicleSort.tPos.spacing;
	VehicleSort.bgW = getTextWidth(size, txt) + VehicleSort.tPos.padSides;		--Background width will be dynamically adjusted later on. Just a value to get started with
  
	-- We render the heading seperately to have it centered, despite the user config for the list
	local headingY = VehicleSort.tPos.y + size + (VehicleSort.tPos.padHeight * 6);
  
	setTextAlignment(VehicleSort.tPos.alignmentC);
	setTextColor(unpack(VehicleSort.tColor.standard));
	renderText(VehicleSort.tPos.x, headingY, size + VehicleSort.tPos.sizeIncr, tostring(txt)); -- x, y, size, txt

	-- Now set the list alignment based on the config
	if VehicleSort.config[20][2] == 1 then
		setTextAlignment(VehicleSort.tPos.alignmentL);
	elseif VehicleSort.config[20][2] == 3 then
		setTextAlignment(VehicleSort.tPos.alignmentR);
	else
		setTextAlignment(VehicleSort.tPos.alignmentC);
	end
 
	--Just used to figure out if we'll have multiple columns, hence we've to loop through the amount of vehicles to get the total height of the table
	--This assumes that there is just one line/vehicle. But it's a rough guess though
	local chk = yPos + size + VehicleSort.tPos.spacing;
	local chkColNum = 1;
	
	--Min distance to the bottom of the screen
	local minY = ((4 * (size + VehicleSort.tPos.spacing)) + VehicleSort.tPos.padHeight);
	
	for i = 1, cnt do --loop through lines to see if there will be multiple columns needed
		if not VehicleSort:isHidden(VehicleSort.Sorted[i]) then
			chk = chk - size - VehicleSort.tPos.spacing;
			if chk < minY then
				chkColNum = chkColNum + 1;
				chk = yPos + size + VehicleSort.tPos.spacing;	--For a new colum we've to reset this
			end
		end
	end

	local colNum = 1;			--For multiple columns this counter gets increased
	--VehicleSort:dp(string.format('chk {%f} | check for chk {%f} | minY {%f}', chk, size + VehicleSort.tPos.spacing + VehicleSort.tPos.padHeight, minY));
	--VehicleSort:dp(string.format('minY {%s} - chk {%s} - chkColNum {%s}', minY, chk, chkColNum));
	
	-- Calc our maxTxtW based on the expected number of columns.
	--As we support just three columns we do the calc based on 3. In case we don't show a infobox we can add the space of one additional column
	local maxTxtW = 0;
	if VehicleSort.config[16][2] then
		maxTxtW = (VehicleSort.tPos.columnWidth - VehicleSort.tPos.padSides) * (3 / chkColNum);
	else
		maxTxtW = (VehicleSort.tPos.columnWidth - VehicleSort.tPos.padSides) * (3 / chkColNum) + VehicleSort.tPos.columnWidth;
	end
	
	--VehicleSort:dp(string.format('columnWidth {%s} - maxTxtW {%s} - chkColNum {%s}', VehicleSort.tPos.columnWidth, maxTxtW, chkColNum));
	
	for i = 1, cnt do
		local realId = VehicleSort.Sorted[i];
		if not VehicleSort:isHidden(realId) then

			local clr = VehicleSort:getTextColor(i, realId);
			local fullNameTable = VehicleSort:getFullVehicleName(realId);
			txt = table.concat(fullNameTable);

			-- Check if the line is not longer as our max txtLength, otherwise split it up to multiple lines
			-- We don't have to care about the MaxNumImplements anymore, as we already consider that in getFullVehicleName
			local multiLine = {};
			if getTextWidth(size, txt) > maxTxtW then
				while #fullNameTable > 0 do
					local line = ''
					while (fullNameTable[1] ~= nil) and (getTextWidth(size, line) < maxTxtW) do
						line = line .. fullNameTable[1];
						table.remove(fullNameTable, 1);
					end
					table.insert(multiLine, line);
				end
				-- Otherwise we add the lines twice
				txt = '';
			end
			
			bold = VehicleSort:isControlled(realId) and (not g_currentMission.missionDynamicInfo.isMultiplayer or VehicleSort:getControllerName(realId) == g_gameSettings.nickname);
			
			if string.len(txt) > 0 then
				table.insert(texts, {colNum, yPos, size, bold, clr, txt});
				yPos = yPos - size - VehicleSort.tPos.spacing;
				
				-- To find our proper background width and the position of the columns we've to keep track of the longest text
				VehicleSort.bgW = math.max(VehicleSort.bgW, getTextWidth(size, txt) + VehicleSort.tPos.padSides);
			end
			
			-- And for multiline entries we've to add multiple entries to our texts table
			if #multiLine > 0 then
				for k, v in ipairs(multiLine) do
					if string.len(v) > 0 then
						table.insert(texts, {colNum, yPos, size, bold, clr, v});
						yPos = yPos - size - VehicleSort.tPos.spacing;
						VehicleSort.bgW = math.max(VehicleSort.bgW, getTextWidth(size, v) + VehicleSort.tPos.padSides);
					end
				end
			end
		
			
			-- We don't want our background go further than necessary
			bgPosY = math.min(bgPosY, yPos);
		end
		
		if yPos < minY then -- getting near bottom of screen, start a new column
			yPos = VehicleSort.tPos.y;
			colNum = colNum + 1;
		end
	end

	setTextBold(false);

	--Drawing our background
	bgPosY = bgPosY - VehicleSort.tPos.spacing; -- bottom padding

	VehicleSort.bgY = bgPosY;
	VehicleSort.bgH = (y - bgPosY) + size + VehicleSort.tPos.sizeIncr + VehicleSort.tPos.yOffset + VehicleSort.tPos.spacing;
	VehicleSort.bgW = VehicleSort.bgW * colNum;
	if VehicleSort.bgY ~= nil and VehicleSort.bgW ~=nil and VehicleSort.bgH ~= nil then
		VehicleSort:renderBg(VehicleSort.bgX, VehicleSort.bgY, VehicleSort.bgW, VehicleSort.bgH);
	end

	--We've to calculate our X based on each column.
	local tblColWidth = VehicleSort.bgW / colNum;
	
	local tPosXAligned = VehicleSort.tPos.x;
	if VehicleSort.config[20][2] == 1 then
		tPosXAligned = VehicleSort.tPos.x - (tblColWidth / 2) + VehicleSort.tPos.padSides;
	elseif VehicleSort.config[20][2] == 3 then
		tPosXAligned = VehicleSort.tPos.x + (tblColWidth / 2) - VehicleSort.tPos.padSides;
	end
	
	local colX = {};
	colX[0] = tPosXAligned;
	if colNum == 2 then
		colX[1] = tPosXAligned - (tblColWidth / 2);
		colX[2] = tPosXAligned + (tblColWidth / 2);
	elseif colNum == 3 then
		colX[1] = tPosXAligned - tblColWidth;
		colX[2] = tPosXAligned;
		colX[3] = tPosXAligned + tblColWidth;
	else
		--We just support 3 columsn. So this case is primarily for one column and as a 'catch all' which won't work but I don't care for now
		colX[1] = tPosXAligned;
	end	
	
	--VehicleSort:dp(colX, 'drawList');
	
	for k, v in ipairs(texts) do
		if type(v[4]) == 'boolean' then
			setTextBold(v[4]);
		end
		setTextColor(unpack(v[5]));
		local storColNum = v[1];
		--VehicleSort:dp(storColNum, 'drawList', 'storcolNum');
		renderText(colX[storColNum], v[2], v[3], tostring(v[6])); -- x, y, size, txt
	end

	setTextBold(false);
	setTextColor(unpack(VehicleSort.tColor.standard));
  
	if VehicleSort.config[15][2] then
		VehicleSort:drawStoreImage(VehicleSort.Sorted[VehicleSort.selectedIndex]);
	end

	if VehicleSort.config[16][2] then
		VehicleSort:drawInfobox(VehicleSort.Sorted[VehicleSort.selectedIndex])
	end
  
end

function VehicleSort:getVehicles()
	local allveh = g_currentMission.vehicles
	local veh = {}
	
	for k, v in ipairs(allveh) do
		if v.spec_vehicleSort ~= nil then
			v.spec_vehicleSort.realId = k;
			table.insert(veh, v);
		end
	end
	return veh;
end

function VehicleSort:getVehImplements(realId)
	if g_currentMission.vehicles[realId].getAttachedImplements ~= nil then
		if #g_currentMission.vehicles[realId]:getAttachedImplements() > 0 then
			local allImp = {}
			-- Credits to Tardis from FS17
			local function addAllAttached(obj)
				for _, imp in pairs(obj:getAttachedImplements()) do
					addAllAttached(imp.object);
					table.insert(allImp, imp);
				end
			end
                
            addAllAttached(g_currentMission.vehicles[realId]);
			return allImp;
		else
			return nil;
		end
	else
		return nil;
	end
end

-- We can't use getFullName for Attachments as that's causing lua callstacks once CP or a helper is used
-- Hence we build our own full name with the help of the store & brand manager
function VehicleSort:getAttachmentName(obj)
	local val = '';
	if VehicleSort.config[3][2] then
		local brand = VehicleSort:getAttachmentBrand(obj);
		if brand ~= nil then
			val = val .. string.format('%s %s', brand, obj:getName());
		else
			val = val .. string.format('%s ', obj:getName());
		end
	else
		val = val .. string.format('%s', obj:getName());
	end
	--VehicleSort:dp(string.format('val = {%s}', val), getAttachmentName);
	return val;
end

function VehicleSort:getAttachmentBrand(obj)
    local storeItem = g_storeManager:getItemByXMLFilename(obj.configFileName);
    if storeItem ~= nil then
        local brand = g_brandManager:getBrandByIndex(storeItem.brandIndex);
        if brand ~= nil then
            return brand.title;
		else
			return 'Lizard';
        end
    end
end

function VehicleSort:getFillLevel(obj)
	local fillLevel = 0;
	local cap = 0;
	local fillType = "";
	if obj.getFillLevelInformation ~= nil then
		local fillLevelTable = {};
		obj:getFillLevelInformation(fillLevelTable);
		
		for _,fillLevelVehicle in pairs(fillLevelTable) do
			fillLevel = fillLevelVehicle.fillLevel;
			cap = fillLevelVehicle.capacity;
			fillType = g_fillTypeManager.fillTypes[fillLevelVehicle.fillType]['title'];
			--VehicleSort:dp(string.format('FillLevel - realId {%f} - fillLevel {%f} - capacity {%f}', realId, fillLevel, cap), 'getFillLevel');
			--VehicleSort:dp(string.format('fillType {%s} - fillTypeIndex {%s} - filltypeTitle {%s}', fillLevelVehicle.fillType, fillTypeIndex, fillType));			
		end
		
		return fillLevel, cap, fillType;
	end
end

function VehicleSort:getFillDisplay(obj, infoBox)
	local ret = '';
	if VehicleSort.config[6][2] or infoBox then -- Fill-Level-Display active?
		local f, c, t = VehicleSort:getFillLevel(obj);
		
		if not infoBox then t = ""; end;	-- we use the same method for the list and the infobox. But the fillType should just be visible in the infobox
		
		if VehicleSort.config[8][2] or f > 0 then -- Empty should be shown or is not empty
			if c > 0 then -- Capacity more than zero
				if infoBox then  -- show more details in the infobox
					ret = string.format('%d/%d (%d %%) %s', math.floor(f), c, VehicleSort:calcPercentage(f, c), t);
				elseif VehicleSort.config[7][2] then -- Display as percentage
					ret = string.format(' (%d %%) %s', VehicleSort:calcPercentage(f, c), t);
				else -- Display as amount of total capacity
					ret = string.format(' (%d/%d) %s', math.floor(f), c, t);
				end
			end
		end
	end
	
	return ret;
end

function VehicleSort:getFullVehicleName(realId)
	local nam = '';
	local ret = {};
	local fmt = '(%s) ';
  
	if VehicleSort:isParked(realId) then
		nam = '[P] '; -- Prefix for parked (not part of tab list) vehicles
	end
	if (g_currentMission.vehicles[realId].getIsCourseplayDriving ~= nil and g_currentMission.vehicles[realId]:getIsCourseplayDriving()) then -- CoursePlay
		nam = nam .. string.format(fmt, g_i18n.modEnvironments[VehicleSort.ModName].texts.courseplay);
	elseif VehicleSort:isHired(realId) then -- credit: Vehicle Groups Switcher mod
		nam = nam .. string.format(fmt, g_i18n.modEnvironments[VehicleSort.ModName].texts.hired);
	--elseif (veh.modFM ~= nil and veh.modFM.FollowVehicleObj ~= nil) then
	--	nam = nam .. string.format(fmt, g_i18n.modEnvironments[VehicleSort.ModName].texts.followme);
	elseif VehicleSort:isControlled(realId) then
		local con = VehicleSort:getControllerName(realId);
		if VehicleSort.config[5][2] and con ~= nil and con ~= 'Unknown' and con ~= '' then
			nam = nam .. string.format(fmt, con);
		end
	end


	if VehicleSort:isTrain(realId) then
		nam = nam .. VehicleSort:getName(realId, string.format('%s', g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_train));
	elseif VehicleSort:isCrane(realId) then
		nam = nam .. VehicleSort:getName(realId, string.format('%s', g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_crane));
	elseif VehicleSort.config[3][2] then -- Show brand
		nam = nam .. string.format('%s', VehicleSort:getNameBrand(realId));
	else
	  --VehicleSort:dp(veh.spec_vehicleSort, 'getFullVehicleName', 'Table spec_vehicleSort');
	  nam = nam .. string.format('%s', VehicleSort:getName(realId));
	end
	
	-- Show horse power
	if VehicleSort.config[4][2] then
		local horsePower = VehicleSort:getHorsePower(realId);
		if horsePower ~= nil then
			nam = nam .. " (" .. horsePower .. string.format(' %s)', g_i18n.modEnvironments[VehicleSort.ModName].texts.horsePower);
		end
	end

	table.insert(ret, nam .. VehicleSort:getFillDisplay(g_currentMission.vehicles[realId]));

	if VehicleSort:getVehImplements(realId) ~= nil and VehicleSort.config[12][2] then
		local implements = VehicleSort:getVehImplements(realId);
		
		local maxCount = VehicleSort.config[26][2];
		if #implements < maxCount then
			maxCount = #implements;
		end
		
		local linkWord = string.format(' %s', g_i18n.modEnvironments[VehicleSort.ModName].texts.with);
		for i=1, maxCount do
			local imp = implements[i];
			if (imp ~= nil and imp.object ~= nil) then
				if i > 1 then
					linkWord = "&";
				end
				table.insert(ret, string.format('%s %s%s ', linkWord, VehicleSort:getAttachmentName(imp.object), VehicleSort:getFillDisplay(imp.object)));		
			end
		end
	end

	return ret;
end

function VehicleSort:getName(realId, sFallback)
	nam = g_currentMission.vehicles[realId]:getName();
	if nam == nil then
		nam = obj.typeName;
	end	
	if nam == nil or nam == '' then
		return sFallback;
	else
		return nam;
	end
end

function VehicleSort:getNameBrand(realId)
	--return Utils.getNoNil(getXMLString(obj.xmlFile, 'vehicle.storeData.brand'), 'LIZARD');
	return g_currentMission.vehicles[realId]:getFullName();
end

function VehicleSort:getOrderedVehicles()
	local ordered = {};
	local unordered = {};
	local orderedToOrder = {};
	VehicleSort.HiddenCount = 0;
	local vehList = VehicleSort:getVehicles();
  
	for _, veh in pairs(vehList) do
		if veh.spec_vehicleSort.orderId ~= nil then
			table.insert(orderedToOrder, {orderId=veh.spec_vehicleSort.orderId, realId=veh.spec_vehicleSort.realId} );
		else
			table.insert(unordered, veh.spec_vehicleSort.realId);
		end
		
		-- Keep track of hidden items, so that we're not showing an empty list
		if VehicleSort:isHidden(veh.spec_vehicleSort.realId) then
			VehicleSort.HiddenCount = VehicleSort.HiddenCount + 1;
		end
	end
	
	-- Now order our temp table based on the actual orderId
	table.sort(orderedToOrder, function(a,b) return a['orderId'] < b['orderId'] end)
	-- And to avoid any holes in the order or dups we'll just add them to a new ordered table
	for _, v in ipairs(orderedToOrder) do
		table.insert(ordered, v['realId']);
	end
	
	local cntOrdered = #ordered;
	
	if unordered ~= nil then
		for _, v in pairs(unordered) do
			table.insert(ordered, v);
		end
	end
	
	-- We might have to reorder the list in case we've missing entries or completely new vehicles
	if #vehList ~= cntOrdered or #unordered ~= 0 then
		VehicleSort:dp(string.format('Reshuffle of vehicles required. #vehList {%d} - cntOrdered {%d} - #unordered {%d}', #vehList, cntOrdered, #unordered));
		ordered = VehicleSort:reshuffleVehicles(ordered);
	end
	
	VehicleSort:SyncSorted();
	return ordered;

end

function VehicleSort:reshuffleVehicles(list)
	local newList = {};
	local i = 1;
	for k, v in ipairs(list) do
		table.insert(newList, v);
		if k ~= i then
			g_currentMission.vehicles[v]['spec_vehicleSort']['orderId'] = i;
		end
		i = i + 1;
	end

	return newList;
end

function VehicleSort:getTextColor(index, realId)
	--VehicleSort:dp(veh, 'getTextColor');
	if index == VehicleSort.selectedIndex then
		if VehicleSort.selectedLock then
			return VehicleSort.tColor.locked;
		else
			return VehicleSort.tColor.selected;
		end
	elseif VehicleSort:isParked(realId) then
		return VehicleSort.tColor.isParked;
	elseif VehicleSort:isControlled(realId) then
		return VehicleSort.tColor.self;
	elseif VehicleSort:isHired(realId) then
		return VehicleSort.tColor.hired;
	elseif VehicleStatus:getIsMotorStarted(g_currentMission.vehicles[realId]) then
		return VehicleSort.tColor.motorOn;
--  FollowMe is not out there yet
--  elseif (veh.modFM ~= nil and veh.modFM.FollowVehicleObj ~= nil) then
--	return VehicleSort.tColor.followme
	else
		if VehicleSort.config[27][2] then						--Alterate the list colors if enabled
			if index % 2 == 0 then
				return VehicleSort.tColor.standard;
			else
				return VehicleSort.tColor.standard2;
			end
		else
				return VehicleSort.tColor.standard;
		end
	end
end

function VehicleSort:getTextSize()
  local val = tonumber(VehicleSort.config[9][2]);
  if val == nil or val < 1 or val > 3 then
    val = 2;
  end
  if val == 1 then
    return VehicleSort.tPos.sizeSmall;
  elseif val == 3 then
    return VehicleSort.tPos.sizeBig;
  else
    VehicleSort.config[9][2] = 2;
    return VehicleSort.tPos.size;
  end
end

function VehicleSort:getHorsePower(realId)
	if VehicleSort:isTrain(realId) then
		--VehicleSort:dp(string.format('isTrain -> realId {%s}', tostring(realId)), 'getHorsePower');
		--return VehicleSort:getHorsePowerFromStore(realId);
	else
		local veh = g_currentMission.vehicles[realId];
		if veh.spec_motorized ~= nil then
			local maxMotorTorque = veh.spec_motorized.motor.peakMotorTorque;
			local maxRpm = veh.spec_motorized.motor.maxRpm;
			return math.ceil(maxMotorTorque / 0.0044);
		end
	end
end


--
-- For some reason the g_storemanager behaves funny for trains and never returns a table. so skip this for now
--
function VehicleSort:getHorsePowerFromStore(realId)
	VehicleSort:dp(string.format('realId {%s}', tostring(realId)));
	local confFile = string.lower(g_currentMission.vehicles[realId]['configFileName']);
	--local storeItem = g_storeManager:getItemByXMLFilename(confFile);
	storeItem = g_storeManager.xmlFilenameToItem[confFile:lower()];
	VehicleSort:dp(storeItem);
	if storeItem ~= nil then
		if storeItem.configurations ~= nil then
			VehicleSort:dp(storeItem.configurations);
			if storeItem.configurations.motor ~= nil then
				if storeItem.configurations.motor.power ~= nil then
					return storeItem.configurations.motor.power;
				end
			end
		end
	end
end

function VehicleSort:getControllerName(realId)
	if not VehicleSort:isHired(realId) then
		if g_currentMission.vehicles[realId].getControllerName ~= nil then
			return g_currentMission.vehicles[realId]:getControllerName();
		else
			return "Unknown Controller";
		end
	end
end

function VehicleSort:initVS()
  VehicleSort:dp('Start Init', 'VehicleSort:init');
  if g_dedicatedServerInfo ~= nil then -- Dedicated server does not need the initialization process
    VehicleSort:dp('Skipping undesired initialization on dedicated server.', 'VehicleSort:init');
    return;
  end
  VehicleSort.dbgX = 0.01;
  VehicleSort.dbgY = 0.5;
  VehicleSort.tPos = {};
  VehicleSort.tPos.x = 0.5;
  VehicleSort.tPos.center = 0.5;
  VehicleSort.tPos.y = g_currentMission.inGameMenu.hud.topNotification.origY + g_currentMission.inGameMenu.hud.topNotification.infoOffsetY;  -- y Position of Textfield, originally hardcoded 0.9
  VehicleSort.tPos.yOffset = g_currentMission.inGameMenu.hud.topNotification.infoOffsetY;  --* 1.5; -- y Position offset for headings, originally hardcoded 0.007
  VehicleSort.tPos.size = g_currentMission.inGameMenu.hud.topNotification.infoTextSize;  -- TextSize, originally hardcoded 0.018
  VehicleSort.tPos.sizeBig = VehicleSort.tPos.size + 0.002;
  VehicleSort.tPos.sizeSmall = VehicleSort.tPos.size - 0.002;
  VehicleSort.tPos.sizeIncr = g_currentMission.inGameMenu.hud.speedMeter.cruiseControlTextOffsetY; -- Text size increase for headings
  VehicleSort.tPos.spacing = g_currentMission.inGameMenu.hud.speedMeter.cruiseControlTextOffsetY;  -- Spacing between lines, originally hardcoded 0.005
  VehicleSort.tPos.padHeight = VehicleSort.tPos.spacing;
  VehicleSort.tPos.padSides = VehicleSort.tPos.padHeight;
  --VehicleSort.tPos.columnWidth = (((1 - VehicleSort.tPos.x) / 2) );
  VehicleSort.tPos.columnWidth = 1 / 5;				--Max 3 columns for vehicles, and one left/right for spacing
  VehicleSort.tPos.alignmentL = RenderText.ALIGN_LEFT;  -- Text Alignment
  VehicleSort.tPos.alignmentC = RenderText.ALIGN_CENTER;  -- Text Alignment
  VehicleSort.tPos.alignmentR = RenderText.ALIGN_RIGHT;  -- Text Alignment
  
  if g_seasons ~= nil then
    VehicleSort:dp('Seasons mod detected. Lowering VehicleSort display to below the seasons weather display to avoid overlap', 'VehicleSort:init');
    VehicleSort.tPos.y = VehicleSort.tPos.y - (6 * VehicleSort.tPos.size) - (6 * VehicleSort.tPos.spacing);
  end
  VehicleSort:dp(VehicleSort.tPos, 'VehicleSort:init', 'tPos');
  VehicleSort.userPath = getUserProfileAppPath();
  VehicleSort.saveBasePath = VehicleSort.userPath .. 'modsSettings/VehicleExplorer/';
  if g_currentMission.missionDynamicInfo.serverAddress ~= nil then --multi-player game and player is not the host (dedi already handled above)
    VehicleSort.savePath = VehicleSort.saveBasePath .. g_currentMission.missionDynamicInfo.serverAddress .. '/';
  else
    VehicleSort.savePath = VehicleSort.saveBasePath .. 'savegame' .. g_careerScreen.selectedIndex .. '/';
  end
  createFolder(VehicleSort.userPath .. 'modsSettings/');
  createFolder(VehicleSort.saveBasePath);
  createFolder(VehicleSort.savePath);
  VehicleSort.xmlFilename = VehicleSort.savePath .. 'VeExConfig.xml';
  VehicleSort.bg = createImageOverlay('dataS2/menu/blank.png'); --credit: Decker_MMIV, VehicleGroupsSwitcher mod
  VehicleSort.bgX = 0.5;
  
  VehicleSort:dp(string.format('Initialized userPath [%s] saveBasePath [%s] savePath [%s]',
    tostring(VehicleSort.userPath),
    tostring(VehicleSort.saveBasePath),
    tostring(VehicleSort.savePath)), 'VehicleSort:init');
end

function VehicleSort:isCrane(realId)
	return g_currentMission.vehicles[realId]['typeName'] == 'crane';
end

function VehicleSort:isHidden(realId)
	return (VehicleSort:isTrain(realId) and not VehicleSort.config[1][2]) or (VehicleSort:isCrane(realId) and not VehicleSort.config[2][2]) or (VehicleSort:isSteerableImplement(realId) and not VehicleSort.config[11][2]);
end

function VehicleSort:isTrain(realId)
	--VehicleSort:dp(string.format('realId {%d}', realId), 'isTrain');
	return g_currentMission.vehicles[realId]['typeName'] == 'locomotive';
end

function VehicleSort:isSteerableImplement(realId)
	return g_currentMission.vehicles[realId]['spec_attachable'] ~= nil;
end

function VehicleSort:isControlled(realId)
	if not VehicleSort:isHired(realId) and g_currentMission.vehicles[realId].getIsControlled ~= nil then
		return g_currentMission.vehicles[realId]:getIsControlled(); 
	end
end

function VehicleSort:isParked(realId)
	if g_currentMission.vehicles[realId] ~= nil and g_currentMission.vehicles[realId].getIsTabbable ~= nil then
		return not g_currentMission.vehicles[realId]:getIsTabbable();
	end
end

-- ToDo
function VehicleSort:isHired(realId)
	if g_currentMission.vehicles[realId].spec_aiVehicle ~= nil then
		return g_currentMission.vehicles[realId].spec_aiVehicle.isActive;
	end
end

function VehicleSort:loadConfig()
	if fileExists(VehicleSort.xmlFilename) then
		VehicleSort.saveFile = loadXMLFile('VehicleSort.loadFile', VehicleSort.xmlFilename);

		if hasXMLProperty(VehicleSort.saveFile, VehicleSort.keyCon) then

			VehicleSort:dp('Config file found.', 'VehicleSort:loadConfig');
			for i = 1, #VehicleSort.config do
				if VehicleSort:contains({9},i) then				--txtsize as int with max value 3
					local int = getXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. VehicleSort.config[i][1]);
					if tonumber(int) == nil or tonumber(int) <= 0 or tonumber(int) > 3 then
						int = VehicleSort.txtSizeDef;
						print("VeEx Config: Invalid saved value for txtSizeDef. Set default value: " .. VehicleSort.txtSizeDef);
					else
						int = math.floor(tonumber(int));
					end
					VehicleSort.config[i][2] = int;
					VehicleSort:dp(string.format('txtSize value set to [%d]', int), 'VehicleSort:loadConfig');
				elseif VehicleSort:contains({24, 25, 26}, i) then	--max implements info & image, showImplementsMax as int with max value 9
					local int = getXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. VehicleSort.config[i][1]);
					if tonumber(int) == nil or tonumber(int) <= 0 or tonumber(int) > 9 then
						int = VehicleSort.config[i][2];
						print("VeEx Config: Invalid saved value for " .. VehicleSort.config[i][1] .. ". Set default value: " .. tostring(VehicleSort.config[i][2]));
					else
						int = math.floor(tonumber(int));
					end
					VehicleSort.config[i][2] = int;
					VehicleSort:dp(string.format('txtSize value set to [%d]', int), 'VehicleSort:loadConfig');
				elseif VehicleSort:contains({10, 17},i) then
					local flt = getXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. VehicleSort.config[i][1]);
					if tonumber(flt) == nil or tonumber(flt) <= 0 or tonumber(flt) > 1 then
						flt = VehicleSort.config[i][2];
						print("VeEx Config: Invalid saved value for " .. VehicleSort.config[i][1] .. ". Set default value: " .. tostring(VehicleSort.config[i][2]));
					else
						flt = tonumber(string.format('%.1f', tonumber(flt)));
					end
					VehicleSort.config[i][2] = flt;
					VehicleSort:dp(string.format('%s value set to [%f]',tostring(VehicleSort.config[i][1]), flt), 'VehicleSort:loadConfig');
				elseif i == 20 then
					local int = getXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. VehicleSort.config[i][1]);
					if tonumber(int) == nil or tonumber(int) <= 0 or tonumber(int) > 3 then
						int = VehicleSort.listAlignment;
						print("VeEx Config: Invalid saved value for " .. VehicleSort.config[i][1] .. ". Set default value: " .. tostring(VehicleSort.config[i][2]));
					else
						int = math.floor(tonumber(int));
					end
					VehicleSort.config[i][2] = int;
					VehicleSort:dp(string.format('listAlignment value set to [%d]', int), 'VehicleSort:loadConfig');
				else
					local b = getXMLBool(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. VehicleSort.config[i][1]);
					if b ~= nil then
						VehicleSort.config[i][2] = b;
					end
				end
			end
			print("VeExConfig loaded");
		end
	end
end

function VehicleSort:moveDown(moveSpeed)
	if moveSpeed == nil then
		moveSpeed = 1;
	end
	local oldIndex = VehicleSort.selectedIndex;
	VehicleSort.selectedIndex = VehicleSort.selectedIndex + moveSpeed;
	if VehicleSort.selectedIndex > #VehicleSort.Sorted then
		VehicleSort.selectedIndex = 1;
	end
	if VehicleSort:isHidden(VehicleSort.Sorted[VehicleSort.selectedIndex]) then
		VehicleSort:moveDown();
	end
	if VehicleSort.selectedLock then
		VehicleSort:reSort(oldIndex, VehicleSort.selectedIndex);
	end
end

function VehicleSort:moveUp(moveSpeed)
	if moveSpeed == nil then
		moveSpeed = 1;
	end
	local oldIndex = VehicleSort.selectedIndex;
	VehicleSort.selectedIndex = VehicleSort.selectedIndex - moveSpeed;
	if VehicleSort.selectedIndex < 1 then
		VehicleSort.selectedIndex = #VehicleSort.Sorted;
	end
	if VehicleSort:isHidden(VehicleSort.Sorted[VehicleSort.selectedIndex]) then
		VehicleSort:moveUp();
	end
	if VehicleSort.selectedLock then
		VehicleSort:reSort(oldIndex, VehicleSort.selectedIndex);
	end
end

function VehicleSort:moveConfigDown()
	VehicleSort.selectedConfigIndex = VehicleSort.selectedConfigIndex + 1;
	if VehicleSort.selectedConfigIndex > #VehicleSort.config then
		VehicleSort.selectedConfigIndex = 1;
	end
	VehicleSort.selectedRealConfigIndex = VehicleSort.orderedConfig[VehicleSort.selectedConfigIndex][1];
end

function VehicleSort:moveConfigUp()
	VehicleSort.selectedConfigIndex = VehicleSort.selectedConfigIndex - 1;
	if VehicleSort.selectedConfigIndex <= 0 then
		VehicleSort.selectedConfigIndex = #VehicleSort.config;
	end
	VehicleSort.selectedRealConfigIndex = VehicleSort.orderedConfig[VehicleSort.selectedConfigIndex][1];
end

function VehicleSort:renderBg(x, y, w, h)
  setOverlayColor(VehicleSort.bg, 0, 0, 0, VehicleSort.config[10][2]);
  renderOverlay(VehicleSort.bg, x - w / 2, y, w, h);
end

function VehicleSort:reSort(old, new)
	VehicleSort:dp(string.format('reSort old {%d} - new {%d}', old, new), 'reSort()');
	local u = VehicleSort.Sorted[old];
	table.remove(VehicleSort.Sorted, old);
	table.insert(VehicleSort.Sorted, new, u);
	VehicleSort:SyncSorted();
	VehicleSort.dirtyState = true;
end

function VehicleSort:SyncSorted()
	for k, v in ipairs(VehicleSort.Sorted) do
		if g_currentMission.vehicles[v] ~= nil then
			if g_currentMission.vehicles[v]['spec_vehicleSort'] ~= nil then
				if g_currentMission.vehicles[v]['spec_vehicleSort']['id'] ~= g_currentMission.vehicles[v]['id'] then
					g_currentMission.vehicles[v]['spec_vehicleSort']['orderId'] = nil;
				else
					g_currentMission.vehicles[v]['spec_vehicleSort']['orderId'] = k;
					g_currentMission.vehicles[v]['spec_vehicleSort']['realId'] = v;
				end
			end
		else
			-- When there is o vehicle, we can actually drop that entry
			table.remove(VehicleSort.Sorted, k);
		end
	end
end

function VehicleSort:SyncSortedWithGame()
	local allVeh = {}	
	local newOrder = {};
	local newSorted = {};

	for _, v in ipairs(g_currentMission.vehicles) do
		table.insert(allVeh, v);
	end
	
	for k, v in ipairs(VehicleSort.Sorted) do
		VehicleSort:dp(string.format('Sorted Index {%d}, realId {%d}, Vehicle {%s}', k, v, g_currentMission.vehicles[v]['configFileName']), 'SyncSortedWithGame');
		
		local newVeh = g_currentMission.vehicles[v];
		newVeh.spec_vehicleSort.orderId = k;
		newVeh.spec_vehicleSort.realId = k;
		table.insert(newOrder, newVeh);
		table.insert(newSorted, k);
	end
	-- Add the unsorted vehicles like trailers, implements etc.
	for k, _ in pairs(allVeh) do
		if allVeh[k]['spec_vehicleSort'] == nil then
			table.insert(newOrder, allVeh[k]);
		end
	end

	VehicleSort:dp(string.format('#newOrder {%d} - #g_currentMission.vehicles {%d}', #newOrder, #g_currentMission.vehicles), 'SyncSortedWithGame');
	if #newOrder == #g_currentMission.vehicles then
		VehicleSort.Sorted = newSorted;
		g_currentMission.vehicles = newOrder;
		-- Update tab order
		--g_inputBinding.events['SWITCH_VEHICLE'].targetObject.loadVehiclesById = newOrder;
		--g_inputBinding.events['SWITCH_VEHICLE_BACK'].targetObject.loadVehiclesById = newOrder;
		
		VehicleSort:dp('Write back of orderd vehicles to g_currentMission.vehicles');
	end
end

function VehicleSort:toggleParkState(selectedIndex)
	local realId = VehicleSort.Sorted[selectedIndex];
	local parked = not g_currentMission.vehicles[realId]:getIsTabbable();
	if parked then
		g_currentMission.vehicles[realId]:setIsTabbable(true);
	else
		g_currentMission.vehicles[realId]:setIsTabbable(false);
	end
	VehicleSort:dp(string.format('realId {%d} - parked {%s}', realId, tostring(parked)), 'VehicleSort:toggleParkState');
end

function VehicleSort:saveConfig()
	VehicleSort.saveFile = createXMLFile('VehicleSort.saveFile', VehicleSort.xmlFilename, VehicleSort.keyCon);
	for i = 1, #VehicleSort.config do
		if VehicleSort:contains({9, 20, 24, 25, 26}, i) then		-- int values
			setXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. tostring(VehicleSort.config[i][1]), tostring(VehicleSort.config[i][2]));
		elseif VehicleSort:contains({10, 17}, i) then				-- floats
			setXMLString(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. tostring(VehicleSort.config[i][1]), string.format('%.1f', VehicleSort.config[i][2]));
		else
			setXMLBool(VehicleSort.saveFile, VehicleSort.keyCon .. '.' .. tostring(VehicleSort.config[i][1]), VehicleSort.config[i][2]);
		end
	end
	saveXMLFile(VehicleSort.saveFile);

  print("VehicleSort config saved");
end

function VehicleSort:drawStoreImage(realId)
	if not VehicleSort:isTrain(realId) and not VehicleSort:isCrane(realId) then
		local imgFileName = VehicleSort:getStoreImageByConf(g_currentMission.vehicles[realId]['configFileName']);
		--VehicleSort:dp(string.format('configFileName {%s}', configFileName));
		--VehicleSort:dp(storeItem, 'drawStoreImage');

		local storeImage = createImageOverlay(imgFileName);
		if storeImage > 0 then
			local storeImgX, storeImgY = getNormalizedScreenValues(128, 128)
			local imgX = 0.5 - VehicleSort.bgW / 2 - storeImgX;
			local imgY = VehicleSort.config[17][2] - storeImgY;

			
			-- Background rendering for the images, based on the saved configvalue
			if VehicleSort.config[19][2] then
				local bgW = storeImgX;
				local bgH = storeImgY;
				local bgX = imgX + (bgW / 2);
				local bgY = imgY;
				VehicleSort:renderBg(bgX, bgY, bgW, bgH);
			end
			-- Must be rendered after the background, otherwise it's covered by it
			renderOverlay(storeImage, imgX, imgY, storeImgX, storeImgY)
			
			if (VehicleSort:getVehImplements(realId) ~= nil) and (imgFileName ~= "data/store/store_empty.png") then
				local impList = VehicleSort:getVehImplements(realId);
				for i = 1, VehicleSort.config[24][2] do			-- Limit to the configured amount of implements to show
					local imp = impList[i];
					if imp ~= nil and imp.object ~= nil then
						local imgFileName = VehicleSort:getStoreImageByConf(imp.object.configFileName);
						local storeImage = createImageOverlay(imgFileName);
						if storeImage > 0 then
							local imgY = VehicleSort.config[17][2] - (storeImgY * (i + 1) );
							
							
							-- Background rendering for the images, based on the saved configvalue
							if VehicleSort.config[19][2] then
								local bgW = storeImgX;
								local bgH = storeImgY;
								local bgX = imgX + (bgW / 2);
								local bgY = imgY;
								VehicleSort:renderBg(bgX, bgY, bgW, bgH);
							end
							-- Must be rendered after the background, otherwise it's covered by it
							
							renderOverlay(storeImage, imgX, imgY, storeImgX, storeImgY)
						end
					end
				end
			end
		end
	end
end

function VehicleSort:getStoreImageByConf(confFile)
	local storeItem = g_storeManager.xmlFilenameToItem[string.lower(confFile)];
	if storeItem ~= nil then
		local imgFileName = storeItem.imageFilename;
		--if imgFileName == 'data/vehicles/train/locomotive01/store_locomotive01.png' or imgFileName == 'data/vehicles/train/locomotive04/store_locomotive04.png' then
		if string.find(imgFileName, 'locomotive') then
			--imgFileName = Utils.getFilename('img/train.png', VehicleSort.ModDirectory);
			imgFileName = "data/store/store_empty.png";
		end
		return imgFileName;
	end
end

function VehicleSort:drawInfobox(realId)
	if realId ~= nil then
	
		local textTable = VehicleSort:getInfoTexts(realId);
	
		setTextAlignment(VehicleSort.tPos.alignmentR);
		setTextColor(unpack(VehicleSort.tColor.standard));
		local txtSize = VehicleSort.tPos.sizeSmall;
		local imgWidth, _ = getNormalizedScreenValues(128,128);
		if not VehicleSort.config[15][2] then				-- If there is no picture we can move more right
			imgWidth = 0.01;
		end
		local infoX = 0.5 - VehicleSort.bgW / 2 - imgWidth - VehicleSort.tPos.padSides;
		local infoY = VehicleSort.config[17][2];
		local txtY = infoY - VehicleSort.tPos.padHeight - txtSize - VehicleSort.tPos.spacing;
		local txtWidth = getTextWidth(txtSize, "Info");
		
		local texts = {};
		for _, t in ipairs(textTable) do
			local tWidth = getTextWidth(txtSize, t);
			renderText(infoX, txtY, txtSize, tostring(t));
			txtY = txtY - txtSize - VehicleSort.tPos.spacing;
			txtWidth = math.max(txtWidth, tWidth);
		end
		
		setTextAlignment(VehicleSort.tPos.alignmentL);		
		
		-- Background rendering for the infobox, based on the saved configvalue
		if VehicleSort.config[18][2] then
			local bgW = txtWidth + (VehicleSort.tPos.padSides * 2);
			-- We have to compensate for the last txtY change in the loop
			local bgH = (txtSize * (#textTable + 1)) + (VehicleSort.tPos.spacing * (#textTable + 1)) + (VehicleSort.tPos.padHeight * 1);
			local bgX = infoX - (bgW / 2) + VehicleSort.tPos.padSides;
			local bgY = txtY;
			VehicleSort:renderBg(bgX, bgY, bgW, bgH);
		end
	end
end

function VehicleSort:getInfoTexts(realId)
	local veh = g_currentMission.vehicles[realId];
	
	if veh ~= nil then
		local texts = {};
		local line = "";
		
		local doSpacing = false;
		
		if (veh.getIsCourseplayDriving ~= nil and veh:getIsCourseplayDriving()) then
			local courseName = "";
			if veh.cp.currentCourseName ~= nil then
				courseName = veh.cp.currentCourseName;
			elseif veh.cp.mapHotspot.fullViewName ~= nil then
				local str = tostring(veh.cp.mapHotspot.fullViewName);
				local t = {}
				for s in str:gmatch("[^\r\n]+") do
					table.insert(t, s)
				end
				if string.len(tostring(t[3])) > 1 then
					courseName = tostring(t[3]);
				end
			else
				courseName = g_i18n.modEnvironments[VehicleSort.ModName].texts.cp_courseunknown
			end
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.cp_course .. ": " .. courseName;
			table.insert(texts, line);
			doSpacing = true;
		end
		
		if VehicleSort:isHired(realId) then
			if veh.getCurrentHelper ~= nil and veh:getCurrentHelper() ~= nil then
				--TODO: g_i18n:getText("ui_helper")
				line = g_i18n.modEnvironments[VehicleSort.ModName].texts.helper .. ": " .. veh:getCurrentHelper().name;
			else
				line = g_i18n.modEnvironments[VehicleSort.ModName].texts.helper .. ": Unknown Helper";
			end
			table.insert(texts, line);
			if not doSpacing then
				doSpacing = true;
			end
		end		

		-- Some spacing, but just if we actually had some data so far
		if doSpacing then
			table.insert(texts, " ");
			doSpacing = false;
		end
		
		-- Get vehicle wear
		if veh.getVehicleDamage ~= nil then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.damage .. ": " .. VehicleSort:calcPercentage(veh:getVehicleDamage(), 1) .. " %";
			table.insert(texts, line);

			if VehicleSort:getVehImplements(realId) ~= nil then
				local impDamage = VehicleStatus:getVehImplementsDamage(realId);
				if #impDamage > 0 then
					for i=1, VehicleSort.config[25][2] do
						table.insert(texts, impDamage[i]);
					end
				end
				doSpacing = true;
			end
			doSpacing = true;
		end
		
		-- Some spacing, but just if we actually had some data so far
		if doSpacing then
			table.insert(texts, " ");
			doSpacing = false;
		end
		
		-- Get vehicle Dirt
		local dirtPerc = VehicleStatus:getDirtPercForObject(g_currentMission.vehicles[realId]);
		if dirtPerc ~= nil then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.dirt .. ": " .. dirtPerc .. " %";
			table.insert(texts, line);		
			doSpacing = true;			
		end
		
		if VehicleSort:getVehImplements(realId) ~= nil then		
			local impDirt = VehicleStatus:getVehImplementsDirt(realId);
			if #impDirt > 0 then
				for i=1, VehicleSort.config[25][2] do
					table.insert(texts, impDirt[i]);
				end
			end
			doSpacing = true;
		end		
		
		-- Some spacing, but just if we actually had some data so far
		if doSpacing then
			table.insert(texts, " ");
			doSpacing = false;
		end

		-- Get vehicle filllevel
		if string.len(VehicleSort:getFillDisplay(veh, true)) > 1 then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.fillLevel .. ":" .. VehicleSort:getFillDisplay(veh, true);
			table.insert(texts, line);
			doSpacing = true;
		end
		
		if VehicleSort:getVehImplements(realId) ~= nil then
			local impFill = VehicleSort:getVehImplementsFillInfobox(realId);
			if #impFill > 0 then
				for i=1, VehicleSort.config[25][2] do
					table.insert(texts, impFill[i]);
				end
				if not doSpacing then
					doSpacing = true;
				end
			end
		end

		-- Some spacing, but just if we actually had some data so far
		if doSpacing then
			table.insert(texts, " ");
			doSpacing = false;
		end
		
		if (not VehicleSort:isTrain(realId)) or (not VehicleSort:isCrane(realId)) then
			-- Diesel & DEF FillLevel
			local level, capacity = VehicleStatus:getDieselLevel(realId)
			if level ~= nil and capacity ~= nil then
				line = g_i18n.modEnvironments[VehicleSort.ModName].texts.diesel .. ": " .. level .. "/" .. capacity .. " " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.liter;
				table.insert(texts, line);
			end
			
			local level, capacity = VehicleStatus:getDefLevel(realId)
			if level ~= nil and capacity ~= nil then
				line = g_i18n.modEnvironments[VehicleSort.ModName].texts.def .. ": " .. level .. "/" .. capacity .. " " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.liter;
				table.insert(texts, line);
			end
		end
		
		-- Speed
		line = g_i18n.modEnvironments[VehicleSort.ModName].texts.speed .. ": " .. VehicleStatus:getSpeedStr(veh);
		table.insert(texts, line);
		
		-- Some spacing
		table.insert(texts, " ");
		
		-- Motor on/TurnedOn, Lights
		if VehicleStatus:getIsMotorStarted(veh) then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.motor .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_on;
		else
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.motor .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_off;
		end
		table.insert(texts, line);
		
		if VehicleStatus:getIsTurnedOn(veh) then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.turnedOn .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_on;
		else
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.turnedOn .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_off;
		end
		table.insert(texts, line);		
		
		if VehicleStatus:getIsLightTurnedOn(veh) then
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.lights .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_on;
		else
			line = g_i18n.modEnvironments[VehicleSort.ModName].texts.lights .. ": " .. g_i18n.modEnvironments[VehicleSort.ModName].texts.vs_off;
		end
		table.insert(texts, line);
		
		return texts;
	end
end



function VehicleSort:getVehImplementsFillInfobox(realId)
	local texts = {};
	local line = "";

	local implements = VehicleSort:getVehImplements(realId);
	if implements ~= nil then
		for i = 1, #implements do
			local imp = implements[i];
			
			if imp ~= nil and imp.object ~= nil and (string.len(VehicleSort:getFillDisplay(imp.object)) > 1) then
				line = g_i18n.modEnvironments[VehicleSort.ModName].texts.fillLevel .. " (" .. string.gsub(VehicleSort:getAttachmentName(imp.object), "%s$", "") .. "): " .. VehicleSort:getFillDisplay(imp.object, true);
				table.insert(texts, line);
			end
		end
	end

	return texts;
end

function VehicleSort:isActionAllowed()
	if g_currentMission.inGameMenu.isOpen or g_currentMission.shopMenu.isOpen then
		return false;
	elseif VehicleSort.showConfig or VehicleSort.showSteerables then
		return true;
	end
end

function VehicleSort:showNoVehicles()
	g_currentMission:showBlinkingWarning(g_i18n.modEnvironments[VehicleSort.ModName].texts.warningNoVehicles, 8000);
end

function VehicleSort:contains(haystack, needle)
	for _, value in pairs(haystack) do
		if value == needle then
			return true;
		end
	end
	return false;
end

function VehicleSort:tabVehicle(backwards)
	-- TBD Not sure yet if it's a good idea to get a ordered List for every tab. Will see if this has any negative
	-- performance impact. If so, I'll just drop it again and define it as 'by design'. Opening/closing the vehList shouldn't be a big deal
	-- especially as it's just an issue when new vehicles get bought/sold
	--if #VehicleSort.Sorted == 0 then
	if #VehicleSort.Sorted == 0 or #VehicleSort.Sorted ~= #VehicleSort:getOrderedVehicles() then
		VehicleSort.Sorted = VehicleSort:getOrderedVehicles();
	end
	
	-- Show the warning in case we've still no vehicles
	if #VehicleSort.Sorted == 0 then
		VehicleSort:showNoVehicles();
	end
	
	if g_currentMission.controlledVehicle == nil then
		conId = nil;
		nextId = 1;
	else
		conId = g_currentMission.controlledVehicle.spec_vehicleSort.orderId

		VehicleSort:getNextInTabList(conId, backwards);
	end
	
	VehicleSort:dp(string.format('conId {%s} - nextId {%d}', tostring(conId), nextId), 'tabVehicle');

	-- We need the loop to check which vehicle we can actually enter
	local run = 1;
	while g_currentMission.vehicles[(VehicleSort.Sorted[nextId])]:getIsControlled() or VehicleSort:isParked(VehicleSort.Sorted[nextId]) do

		VehicleSort:getNextInTabList(nextId, backwards)

		if run == #VehicleSort.Sorted then
			VehicleSort.showNoVehicles();
			return false;
		end
		run = run + 1;
	end
	realVeh = g_currentMission.vehicles[VehicleSort.Sorted[nextId]];
	g_currentMission:requestToEnterVehicle(realVeh);
	
end

function VehicleSort:getNextInTabList(orderId, backwards)
	if backwards then
		if orderId == 1 then
			nextId = #VehicleSort.Sorted;
		else
			nextId = orderId - 1;
		end			
	else
		if orderId == #VehicleSort.Sorted then
			nextId = 1;
		else
			nextId = orderId + 1;
		end
	end
	return nextId;
end

--
-- Extends default game functions
-- This is required to block the camera zoom & handtool selection while drawlist or drawconfig is open
--
function VehicleSort.onInputCycleHandTool(self, superFunc, _, _, direction)
	if not VehicleSort.showSteerables and not VehicleSort.showConfig then
		superFunc(self, _, _, direction);
	end
end

function VehicleSort.zoomSmoothly(self, superFunc, offset)
	if not VehicleSort.showConfig and not VehicleSort.showSteerables then -- don't zoom camera when mouse wheel is used to scroll displayed list
		superFunc(self, offset);
	end
end

-- I think it's convinient to have the moveup/down fast keys on KEY_1 and KEY_2, but that conflicts with the cruisecontrol
function VehicleSort.setCruiseControlMaxSpeed(self, superFunc, speed)
	if not VehicleSort.showConfig and not VehicleSort.showSteerables then
		superFunc(self, speed);
	end
end

if g_dedicatedServerInfo == nil then
  VehicleCamera.zoomSmoothly = Utils.overwrittenFunction(VehicleCamera.zoomSmoothly, VehicleSort.zoomSmoothly);
  Player.onInputCycleHandTool = Utils.overwrittenFunction(Player.onInputCycleHandTool, VehicleSort.onInputCycleHandTool);
  Drivable.setCruiseControlMaxSpeed = Utils.overwrittenFunction(Drivable.setCruiseControlMaxSpeed, VehicleSort.setCruiseControlMaxSpeed);
end